/* $Id: Command.h,v 1.20 2002/04/17 12:25:49 arianne_rpg Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef ARIANNE_Command
#define ARIANNE_Command

#include "Types.h"
#include "1_Message/Message.h"
#include "3_Manager/StatsStructs.h"

/** This is the invalid type of command */
#define INVALID_COMMAND                0

/** This is a Message to Command Command of Admin type. */
#define MESSAGE_M2C_ADMIN_COMMAND      1
/** This is a Message to Command Command of Action type. */
#define MESSAGE_M2C_ACTION_COMMAND     2
/** This is a Message to Command Command of Perception type. */
#define MESSAGE_M2C_PERCEPTION_COMMAND 3

/** This is a Command to Message Command of Admin type. */
#define MESSAGE_C2M_ADMIN_COMMAND      4
/** This is a Command to Message Command of Action type. */
#define MESSAGE_C2M_ACTION_COMMAND     5
/** This is a Command to Message Command of Perception type. */
#define MESSAGE_C2M_PERCEPTION_COMMAND 6

/** These are the Stats Commands of each Layer. */
#define STAT_L0_COMMAND                7
#define STAT_L1_COMMAND                8
#define STAT_L2_COMMAND                9
#define STAT_L3_COMMAND                10
#define STAT_L4_COMMAND                11
#define STAT_L5_COMMAND                12

/** This command is generated on a New connection event. */
#define NEWCONNECTION_COMMAND          13
/** This command is generated on a close connection event. */
#define CLOSECONNECTION_COMMAND        14


/** This command is sent to make a Manager to stop. */
#define POISON_COMMAND                 15

/** Total number of commands. */
#define NUMBER_OF_COMMANDS             (POISON_COMMAND+1)

/** This is the abstract class for Commands.
    Commands is for the engine what Messages is for the network */
class Command
  {
  protected:
  /** Type of command */
  const Uint8 sType;

  /** Create an instance of Command of the given Type */
  Command(Uint8 Type):sType(Type)
    {}

  public:
  /** Create an instance of Command based on the given Command */
  Command(Command const&b):
    sType(b.sType)
    {}
  /** Destroy this instance of Command */
  virtual ~Command()
    {}

  /** This is the type of the message */
  const char GetType()
    {
    return sType;
    }

  /** Returns a copy of the Command */
  virtual Command *clone() const = 0;
  };

class MessageCommand: public Command
  {
  Message *Mes;

  public:
  MessageCommand(Uint8 Type):
    Command(Type),
    Mes(NULL)
    {}

  MessageCommand(MessageCommand const& b):
    Command(b)
    {
    Mes=b.Mes->clone();
    }

  virtual ~MessageCommand()
    {
    if(Mes) 
      {
      delete Mes;
      Mes=0;
      }
    }

  void SetMessage(Message *mes)
    {
    Mes=mes;
    }
  Message *GetMessage() const
    {
    return Mes;
    }

  virtual Command *clone() const
    {
    return new MessageCommand(*this);
    }
  };

class StatL0Command: public Command
  {
  friend class NetworkServerManager;
  friend class StatsManager;

  TemporalInfoLayer0 tmpInfL0;

  public:
  StatL0Command():Command(STAT_L0_COMMAND) {};
  StatL0Command(StatL0Command const &b):Command(b),tmpInfL0(b.tmpInfL0) {};
  virtual ~StatL0Command() {};

  virtual Command *clone() const
    {
    return new StatL0Command(*this);
    }
  };

class StatL1Command: public Command
  {
  friend class MessageServerManager;
  friend class StatsManager;

  TemporalInfoLayer1 tmpInfL1;

  public:
  StatL1Command():Command(STAT_L1_COMMAND) {};
  StatL1Command(StatL1Command const &b):Command(b),tmpInfL1(b.tmpInfL1) {};
  virtual ~StatL1Command() {};

  virtual Command *clone() const
    {
    return new StatL1Command(*this);
    }
  };

class StatL2Command: public Command
  {
  friend class CommandManager;
  friend class StatsManager;

  TemporalInfoLayer2 tmpInfL2;

  public:
  StatL2Command():Command(STAT_L2_COMMAND) {};
  StatL2Command(StatL2Command const &b):Command(b),tmpInfL2(b.tmpInfL2) {};
  virtual ~StatL2Command() {};

  virtual Command *clone() const
    {
    return new StatL2Command(*this);
    }
  };

class StatL3Command: public Command
  {
  friend class AdminManager;
  friend class StatsManager;

  TemporalInfoLayer3 tmpInfL3;

  public:
  StatL3Command():Command(STAT_L3_COMMAND) {};
  StatL3Command(StatL3Command const &b):Command(b),tmpInfL3(b.tmpInfL3) {};
  virtual ~StatL3Command() {};

  virtual Command *clone() const
    {
    return new StatL3Command(*this);
    }
  };

class StatL4Command: public Command
  {
  friend class StatsManager;

  TemporalInfoLayer4 tmpInfL4;

  public:
  StatL4Command():Command(STAT_L4_COMMAND) {};
  StatL4Command(StatL4Command const &b):Command(b),tmpInfL4(b.tmpInfL4) {};
  virtual ~StatL4Command() {};

  virtual Command *clone() const
    {
    return new StatL4Command(*this);
    }
  };


class CloseConnectionCommand: public Command
  {
  friend class AdminManager;
  friend class NetworkServerManager;
  friend class MessageServerManager;

  Uint32 cID;

  public:
  CloseConnectionCommand():
    Command(CLOSECONNECTION_COMMAND), cID(0)
    {
    };
  CloseConnectionCommand(CloseConnectionCommand const& b):
    Command(b), cID(b.cID)
    {
    };
  virtual ~CloseConnectionCommand() {};

  virtual Command *clone() const
    {
    return new CloseConnectionCommand(*this);
    }
  };

class NewConnectionCommand: public Command
  {
  friend class AdminManager;
  friend class NetworkServerManager;
  friend class MessageServerManager;

  Uint32 cID;
  std::string ipAddr;

  public:
  NewConnectionCommand():
    Command(NEWCONNECTION_COMMAND), cID(0), ipAddr("")
    {
    };
  NewConnectionCommand(NewConnectionCommand const& b):
    Command(b), cID(b.cID), ipAddr(b.ipAddr)
    {
    };
  virtual ~NewConnectionCommand() {};

  virtual Command *clone() const
    {
    return new NewConnectionCommand(*this);
    }
  };

/** This command is used to shutdown managers that receive it. */
class PoisonCommand: public Command
  {
  public:
  PoisonCommand():Command(POISON_COMMAND) {};
  PoisonCommand(PoisonCommand const &b):Command(b) {};
  virtual ~PoisonCommand() {};

  virtual Command *clone() const
    {
    return new PoisonCommand(*this);
    }
  };

#endif
