/***************************************************************************
                          thread.h  -  description
                             -------------------
    begin                : Tue Jul 4 2000
    copyright            : (C) 2000 by Ulrich Eckhardt
    email                : Doomster@knuut.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARIANNE_THREAD_H
#define ARIANNE_THREAD_H

/**Encapsulation of the SDL-library threading facilities:

   Danger: The constructor runs in a different thread than the Thread-object itself,
   do not asssume things like threadID() to work properly right after creating the object!

   The auto-delete function is to just fire up one object and stop caring about it. Possible
   applications are SFX-player threads: just start playing a sound now and forget about when it ends.
   Other thing is the auto-delete function, Im not so sure if this is a)necessary and b)properly
   implemented, YMMV - lets hear it.
		
   To create a thread for some task, derive a class from Thread and supply the thread-function.
    If your Thread can run as it is created (if all necessary data are give to the constructor),
   you can call start() in the constructor.
    Else, make all necessary preparations and then start() the thread.

  *@author Ulrich Eckhardt
  */

#include "Mutex.h"
#include <string>

#ifdef macintosh
 #include <Threads.h>
#else
 #include "SDL_thread.h"
#endif


#ifdef macintosh
pascal void* commonThreadFunction( void* data );
#else
extern "C" int commonThreadFunction(void*);
#endif

#include "Exceptions.h"

class Thread
{
    /** not copyable */
    Thread(Thread const&);
    /** not assignable */
    Thread const& operator=(Thread const&);

    std::string m_ThreadName;

public:
    Thread(std::string threadn);
    virtual ~Thread();

    /** start the thread */
    void Start();
	
#ifdef macintosh
    typedef ThreadID ID;
#else
    typedef Uint32 ID;
#endif
    /** Get the thread identifier for the thread */
    ID threadID() const
    {   return m_ID;    }

    std::string threadName() const
    {   return m_ThreadName;    }

    /** Wait for a thread to finish. The return code for the thread function is placed
    in the area pointed to by 'status', if 'status' is not null. */
    virtual void WaitThread(int *status = 0);

    /** Forcefully kill a thread without worrying about its state(welcome to Memory,Leak&Co Ltd) */
    virtual void KillThread();

    class ThreadException: public Exception
    {
	protected:
	    ThreadException(): Exception() {}
    };

    class ThreadInvalidData: public ThreadException
    {
	public:
	    ThreadInvalidData(std::string const& where = "UNKNOWN"): ThreadException()
	    {
		message = where + " - Thread::commonThreadFunction(void* data): called with data = 0";
	    }
    };
    
protected:
    /** This is the thread function that is called by the framework. You have
    to override this to give the class some functionality. */
    virtual int threadFunction()=0;

private:
#ifdef macintosh
#else
    SDL_Thread* m_SDLThread;
#endif

    // the thread's ID
    ID m_ID;

    /** we have to declare this a friend since it accesses the Thread-object's private parts */	
#ifdef macintosh
    friend pascal void* commonThreadFunction( void* data )
#else
    friend int commonThreadFunction(void*);
#endif
};	


/** This class features a thread that perpetually call its loop() function.
Furthermore, this thread doesnt start acting until you specifically tell it
to and features a stop()-method that resembles the waitThread()-method but 
asks the thread to terminate via a simple bool-variable. */

class LoopThread:
    private Thread
{
public:
    LoopThread(std::string threadn);
    ~LoopThread();

    /** the thread is initially not running, this method starts it or
    resumes it if it was suspended */
    virtual void Start();
    /** this method suspends the thread after the current loop()-call
    returns */
    virtual void Suspend();
    /** this method terminates the thread after the current loop()-call
    returns */
    virtual int Stop();

    /** Wait for a thread to finish. The return code for the thread function is placed
    in the area pointed to by 'status', if 'status' is not null. */
    virtual void Wait(int *status = 0);

    /** returns true if the Thread is running */
    bool isRunning();

    /** Get the thread identifier for the thread */
    ID threadID() const
    {   return Thread::threadID();    }

    /** Get the thread identifier for the thread */
    std::string threadName() const
    {   return Thread::threadName();    }

protected:
    virtual void loop()=0;

private:
    /** to pause the thread */
    Mutex m_suspend;
    /** to sync actions that might affect each other */
    Mutex m_sync;
    volatile bool m_suspended;
    volatile bool m_terminate;
    volatile bool m_running;
    virtual int threadFunction();
    LoopThread(LoopThread const&);
    LoopThread& operator=(LoopThread const&);
};
#endif // ARIANNE_THREAD_H
