/* $Id: RsaKey.cpp,v 1.14 2002/03/07 20:37:38 guidoj Exp $ */
/***************************************************************************
    copyright            : (C) 2001 by Benjamin Lerman
    email                : quisar@quisar.ambre.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_CRYPTO
#ifdef HAVE_OPENSSL

#include <iostream>
#include <fstream>
#include <string>

#include <openssl/objects.h>
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#else
#include <objects.h>
#include <rsa.h>
#include <crypto.h>
#endif

#ifndef OPENSSL_free
#define OPENSSL_free(addr)     CRYPTO_free(addr)
#endif

#include "Crypto/RsaKey.h"
#include "Conversion/Conversion.h"

const std::string RSAKey::n("n:0x");
const std::string RSAKey::e("e:0x");
const std::string RSAKey::d("d:0x");
const std::string RSAKey::p("p:0x");
const std::string RSAKey::q("q:0x");
const std::string RSAKey::dmp1("dmp1:0x");
const std::string RSAKey::dmq1("dmq1:0x");
const std::string RSAKey::iqmp("iqmp:0x");

RSAKey::RSAKey(std::string const& path)
{
 LoadKey(path);
}

bool RSAKey::LoadKey(std::string const& path)
{
    std::ifstream file(path.c_str());
    if(! file)
    {
        key = NULL;
        return false;
    }

    key = RSA_new();

    std::string res;
    while( std::getline(file, res)){
        // ignore empty lines
        if(res.empty())
            continue;
#ifdef BROKEN_STD_BASIC_STRING_COMPARE
/* These comparisons do rely on a broken STL-implementation like the one
from gcc < 3. Standard is that the compare()-memberfunction for a
range has the prototype
    int compare(size_t pos, size_t count, string const& s)const;
If this fails, adapt config.h in the top source-dir accordingly.
*/
        if(res.compare(n, 0, n.size()) == 0)
            BN_hex2bn(&(key->n), res.substr(n.size()).c_str());
        else if(res.compare(e, 0, e.size()) == 0)
            BN_hex2bn(&(key->e), res.substr(e.size()).c_str());
        else if(res.compare(d, 0, d.size()) == 0)
            BN_hex2bn(&(key->d), res.substr(d.size()).c_str());
        else if(res.compare(p, 0, p.size()) == 0)
            BN_hex2bn(&(key->p), res.substr(p.size()).c_str());
        else if(res.compare(q, 0, q.size()) == 0)
            BN_hex2bn(&(key->q), res.substr(q.size()).c_str());
        else if(res.compare(dmp1, 0, dmp1.size()) == 0)
            BN_hex2bn(&(key->dmp1), res.substr(dmp1.size()).c_str());
        else if(res.compare(dmq1, 0, dmq1.size()) == 0)
            BN_hex2bn(&(key->dmq1), res.substr(dmq1.size()).c_str());
        else if(res.compare(iqmp, 0, iqmp.size()) == 0)
            BN_hex2bn(&(key->iqmp), res.substr(iqmp.size()).c_str());
#else
        if(res.compare( 0, n.size(), n) == 0)
            BN_hex2bn(&(key->n), res.substr(n.size()).c_str());
        else if(res.compare( 0, e.size(), e) == 0)
            BN_hex2bn(&(key->e), res.substr(e.size()).c_str());
        else if(res.compare( 0, d.size(), d) == 0)
            BN_hex2bn(&(key->d), res.substr(d.size()).c_str());
        else if(res.compare( 0, p.size(), p) == 0)
            BN_hex2bn(&(key->p), res.substr(p.size()).c_str());
        else if(res.compare( 0, q.size(), q) == 0)
            BN_hex2bn(&(key->q), res.substr(q.size()).c_str());
        else if(res.compare( 0, dmp1.size(), dmp1) == 0)
            BN_hex2bn(&(key->dmp1), res.substr(dmp1.size()).c_str());
        else if(res.compare( 0, dmq1.size(), dmq1) == 0)
            BN_hex2bn(&(key->dmq1), res.substr(dmq1.size()).c_str());
        else if(res.compare( 0, iqmp.size(), iqmp) == 0)
            BN_hex2bn(&(key->iqmp), res.substr(iqmp.size()).c_str());
#endif
    };

    return true;
}

RSAKey::RSAKey(std::string const& n, std::string const& e)
{
    key = RSA_new();

    // try to parse the given strings
    if( BN_hex2bn(&(key->n), n.c_str()) && BN_hex2bn(&(key->e), e.c_str()) )
        return;

    // failure
    // TODO: maybe throw std::bad_alloc here ?
    RSA_free(key);
    key = NULL;
}

RSAKey::RSAKey():
    key(NULL)
{}

RSAKey::~RSAKey()
{
    if(key)
        RSA_free(key);
}

bool RSAKey::SaveKey(std::string const &path) const
{
    char *tmp;

    if(key == NULL)
        return false;
 
    std::ofstream file(path.c_str());
    if(! file)
        return false;

    // write the individual key-components to the file
    if(key->n)
    {
        tmp = BN_bn2hex(key->n);
        file << n << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->e)
    {
        tmp = BN_bn2hex(key->e);
        file << e << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->d)
    {
        tmp = BN_bn2hex(key->d);
        file << d << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->p)
    {
        tmp = BN_bn2hex(key->p);
        file << p << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->q)
    {
        tmp = BN_bn2hex(key->q);
        file << q << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->dmp1)
    {
        tmp = BN_bn2hex(key->dmp1);
        file << dmp1 << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->dmq1)
    {
        tmp = BN_bn2hex(key->dmq1);
        file << dmq1 << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->iqmp)
    {
        tmp = BN_bn2hex(key->iqmp);
        file << iqmp << tmp << "\n";
        OPENSSL_free(tmp);
    }

    return true;
}

bool RSAKey::SavePublicKey(std::string const &path) const
{
    char *tmp;

    if(key == NULL)
        return false;

    std::ofstream file(path.c_str());
    if(! file)
        return false;

    // write the public key-components to the file
    if(key->n)
    {
        tmp = BN_bn2hex(key->n);
        file << n << tmp << "\n";
        OPENSSL_free(tmp);
    }
    if(key->e)
    {
        tmp = BN_bn2hex(key->e);
        file << e << tmp << "\n";
        OPENSSL_free(tmp);
    }

    return true;
}

void RSAKey::GenerateKey(const int nbbits, const int ve)
{
    if(key != NULL)
        RSA_free(key);

    key = RSA_generate_key(nbbits, ve, NULL, NULL);
}

void RSAKey::GenerateKey(const int nbbits)
{
    GenerateKey(nbbits, 65535);
}

void RSAKey::GenerateKey()
{
    GenerateKey(2048, 65535);
}

bool RSAKey::SignMessage(std::vector<unsigned char> const& message, std::vector<unsigned char> &signature) const
{
    std::vector<unsigned char> mysignature;
  
    unsigned int size_signature;

    mysignature.resize(RSA_size(key));
 
    if(key == NULL)
        return false;

    signature.clear();

    if(!RSA_sign(NID_sha1, (unsigned char *) &message[0], message.size(), &mysignature[0], &size_signature, key))
        return false;

    CharToVector(&mysignature[0], size_signature, signature);
    return true;
}

bool RSAKey::SignMessage(std::string const& message, std::vector<unsigned char> &signature) const
{
    std::vector<unsigned char> m;
    StringToVector(message, m);
    return SignMessage(m, signature);
}

bool RSAKey::VerifySignature(std::vector<unsigned char> const& message, std::vector<unsigned char> const &signature) const
{
    if(key == NULL)
        return false;

    return (RSA_verify(NID_sha1, (unsigned char *) &message[0], message.size(), (unsigned char *) &signature[0], signature.size(), key));
}

bool RSAKey::VerifySignature(std::string const& message, std::vector<unsigned char> const &signature) const
{
    std::vector<unsigned char> m;
    StringToVector(message, m);
    return VerifySignature(m, signature);
}

bool RSAKey::GetPublicKey(std::string & n, std::string & e)
{
    char *tmp;

    if(! IsValid())
        return false;

    tmp = BN_bn2hex(key->n);
    if(! tmp)
        return false;
    n = std::string(tmp);
    OPENSSL_free(tmp);

    tmp = BN_bn2hex(key->e);
    if(! tmp)
        return false;
    e = std::string(tmp);
    OPENSSL_free(tmp);

    return true;
}

bool RSAKey::IsValid() const
{
    return(key != NULL);
}

#endif
