/* $Id: AList.h,v 1.16 2002/04/26 13:55:24 doomster Exp $ */
#ifndef ARIANNE_RPG_AList_H
#define ARIANNE_RPG_AList_H

#include "Types.h"
#include "Exceptions.h"
#include <cassert>

#define MAX_ALIST_SIZE (1<<20)
#define KEEP_SORT_LIST 

/** This is a list implementation based on pointers.
    The copy-constructor and the operator= copies also the content of the 
    memory.

	Please STOP using it, move instead to std::list etc.
	
	 */
template <class BaseType>
class AList
  {
  protected:
  // the vector of pointers
  BaseType **list;
  // the amouunt of allocated pointers
  size_t NumMaxCells;
  // the amount of used pointers
  size_t NumCells;

  public:
  class SizeException: public Exception
  {
      public:
      SizeException(std::string const& where): Exception()
      {
          message = where + " - AList: SizeException";
      }
  };

  AList(size_t size=0):
    list(new BaseType*[size]),
    NumMaxCells(size)
    {
    Clear();
    }

  AList(AList const& b)
    {
    NumMaxCells=b.NumMaxCells;
    list=new BaseType*[NumMaxCells];
    Clear();

    for(size_t i=0;i<b.Size();++i)
      {
      Add(b[i]->clone());
      }
    }

  virtual ~AList() 
    {
    Delete();

    delete[] list;
    list=0;
    NumMaxCells=0;
    NumCells=0;
    }

  AList& operator=(AList const& b)
    {
    Clear();
    for(size_t i=0;i<b.Size();++i)
      {
      Add(b[i]->clone());
      }

    return *this;
    }

  bool operator==(AList const& b) const
    {
    // compare size
    if( NumCells != b.NumCells)
        return false;

    // compare elements
    for(size_t i=0;i!=b.Size();++i)
      {
      if(list[i]!=b.list[i])
        {
        return false;
        }
      }

    return true;
    }

  /** This empties the list WITHOUT freeing the memory pointed to,
  useful for initializing the vector. */
  void Clear()
    {
    NumCells=0;
    for(size_t i=0;i!=NumMaxCells;++i)
      {
      list[i]=0;
      }
    }

  /** This empties the list freeing the memory. */
  void Delete()
    {
    for(size_t i=0;i<NumCells;++i)
      {
        delete list[i];
        list[i]=0;
      }
    NumCells=0;
    }

  /** Returns true if the list is empty. */
  bool isEmpty() const
    {
    return NumCells==0;
    }

  void Update(size_t i, BaseType *object)
    {
    assert(object);
    assert(i<NumCells);

    delete list[i];
    list[i]=object;

    }

  /** This method Adds an object to the end of the list */
  void Add(BaseType *object)
    {
    assert(object);

    /* Here comes one of the tricky parts, we do an exponential grow of the buffer if needed. */
    if(NumCells<NumMaxCells)
      {
      list[NumCells]=object;
      ++NumCells;
      return;
      }
    else
      {
      if(NumMaxCells==MAX_ALIST_SIZE)
        {
        throw SizeException(__FILE_AND_LINE__);
        }
      else
        {
        NumMaxCells=NumMaxCells<<1;
        if(0==NumMaxCells)
            NumMaxCells = 16; // reasonable starting-value
        BaseType **newlist=new BaseType*[NumMaxCells];

        for(size_t i=0;i<NumCells;++i)
          {
          newlist[i]=list[i];
          }
        newlist[NumCells]=object;
        ++NumCells;
      
        delete[] list;
        list=newlist;
        }
      }
    }

  /** This method remove an object of the list */
  void Free(size_t pos)
    {
    assert(pos<NumCells);
    assert(list[pos]);

    delete list[pos];

    /* We have now two possibilities, keep order or not */
    #ifdef KEEP_SORT_LIST
      /* Keeping */
      for(size_t i=pos+1;i<NumCells;++i)
        {
        list[i-1]=list[i];
        }
      list[NumCells]=0;
      NumCells--;
    #else
      /* Not keeping */
      list[pos]=list[NumCells-1];
      NumCells--;
    #endif            

    return;
    }
 
  /** This method remove the object from the list and return it */
  BaseType *Remove(size_t pos)
    {
    assert(pos<NumCells);
    assert(list[pos]);

    BaseType *tmp=0;
    tmp=list[pos];

    /* We have now two possibilities, keep order or not */
    #ifdef KEEP_SORT_LIST
      /* Keeping */
      for(size_t i=pos+1;i<NumCells;++i)
        {
        list[i-1]=list[i];
        }
      list[NumCells]=0;
      NumCells--;
    #else
      /* Not keeping */
      list[pos]=list[NumCells-1];
      NumCells--;
    #endif       
   
    return tmp;     
    }

  /** This returns the number of objects on the list. */
  Uint32 Size() const
    {
    return NumCells;
    }

  /** This is operator to access to the elements of the list. */
  BaseType *operator[](size_t i) const
    {
    assert(i<NumCells);

    return list[i];
    }

  /** This operator add the contents of listb to actual list. */
  AList<BaseType> &operator+=(AList<BaseType> &listb)
    {
    for(size_t i=0;i<listb.Size();++i)
      {
      Add(listb[i]);
      }

    return *this;
    }
  };

#endif
