/* $Id: Serializer.h,v 1.21 2002/03/20 11:02:30 arianne_rpg Exp $ */
/***************************************************************************
                          Serializer.h  -  description
                             -------------------
    begin                : Sun Mar 18 2001
    copyright            : (C) 2001 by Ulrich Eckhardt
    email                : Doomster@knuut.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARIANNE_SERIALIZER_H
#define ARIANNE_SERIALIZER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Types.h"
#include "Exceptions.h"

#include <vector>
#include <string>
#include <list>

/**A helper-class to ease serialization of other data-types that are to be sent over the network  */
class Serializer
  {
  // not copyable
  Serializer(Serializer const&);
  // not assignable
  Serializer& operator=(Serializer const&);

  std::vector<Uint8> m_data;
  Uint32 m_index;

  public:
  Serializer();
  Serializer(std::vector<Uint8> const& data);
  ~Serializer();

  class EOFException: public Exception
  {
      public:
      EOFException(std::string const &where = "UNKNOWN"): Exception()
      {
          message = where + " - class Serializer: trying to read beyond end of data";
      }
  };
  /** add a byte */
  void push(Uint8 c);
  
  /** retrieve a byte */
  Uint8 pop();

  /** retrieve the full data vector */
  void tovector(std::vector<Uint8> &data) const;
  };



/**
  The following overloaded operators write or read one of the basic data-types.
  These are designed to do so in an endian-independant way.
**/
Serializer& operator<<(Serializer& s, Uint8 const& t);
Serializer& operator<<(Serializer& s, Uint16 const& t);
Serializer& operator<<(Serializer& s, Uint32 const& t);

Serializer& operator>>(Serializer& s, Uint8& t);
Serializer& operator>>(Serializer& s, Uint16& t);
Serializer& operator>>(Serializer& s, Uint32& t);

inline Serializer& operator<<(Serializer& s, Sint8 t)
  {
  return s << reinterpret_cast<Uint8&>(t);   
  }

inline Serializer& operator<<(Serializer& s, Sint16 t)
  {
  return s << reinterpret_cast<Uint16&>(t);   
  }

inline Serializer& operator<<(Serializer& s, Sint32 t)
  {
  return s << reinterpret_cast<Uint32&>(t);   
  }

 
#ifdef MACOSX

inline Serializer& operator<<(Serializer& s, long unsigned int t)
{
    return s << reinterpret_cast<Uint32&>(t);
}

inline Serializer& operator>>(Serializer& s, long unsigned int t)
{
    return s >> reinterpret_cast<Uint32&>(t);
}

#endif

inline Serializer& operator>>(Serializer& s, Sint8& t)
  {
  return s >> reinterpret_cast<Uint8&>(t);   
  }

inline Serializer& operator>>(Serializer& s, Sint16& t)
  {
  return s >> reinterpret_cast<Uint16&>(t);   
  }

inline Serializer& operator>>(Serializer& s, Sint32& t)
  {
  return s >> reinterpret_cast<Uint32&>(t);   
  }

Serializer& operator<<(Serializer& s, bool const& b);
Serializer& operator>>(Serializer& s, bool& b);

Serializer& operator<<(Serializer& s, std::string const& b);
Serializer& operator<<(Serializer& s, std::string & b);
Serializer& operator>>(Serializer& s, std::string& b);


/** These two templates enable you to simply add a read() and write()-method to your
class and then be able to use insertion(<<) or extraction(>>) operators with it.
*/
template <class T> inline Serializer& operator<<(Serializer&s, T const& t)
    {
    t.write(s);
    return s;
    }

template <class T> inline Serializer& operator>>(Serializer&s, T& t)
    {
    t.read(s);
    return s;
    }


/*
  some more templates for typical STL-containers
*/
#ifndef WIN32
template<class T> inline Serializer& operator<<(Serializer& s, std::list<T> const& l)
{
    s << (Uint32)l.size();
    std::list<T>::const_iterator last = l.end();
    for(std::list<T>::const_iterator it = l.begin(); it!=last; ++it)
        s << *it;

    return s;
}

template<class T> inline Serializer& operator>>(Serializer& s, std::list<T>& l)
{
    Uint32 size;
    s >> size;
    l.resize(size);
    std::list<T>::iterator last = l.end();
    for(std::list<T>::iterator it=l.begin(); it!=last; ++it)
        s >> *it;
    return s;
}
#else
/** Visual C has problems choosing the correct template. */
inline Serializer& operator<<(Serializer& s, std::list<std::string> const& l)
  {
  s << (Uint32)l.size();
  std::list<std::string>::const_iterator last = l.end();
  for(std::list<std::string>::const_iterator it = l.begin(); it!=last; ++it)
    {
    s << (std::string)(*it);
    }
  return s;
  }

inline Serializer& operator>>(Serializer& s, std::list<std::string>& l)
  {
  Uint32 size;
  s >> size;
  l.clear();

  while(size--)
    {
    std::string tmp;
    s >> tmp;
    l.push_back(tmp);
    }

  return s;
  }

#endif

#ifndef WIN32
template<class T> inline Serializer& operator<<(Serializer& s, std::vector<T> const& l)
    {
        s << (Uint32)l.size();
	std::vector<T>::const_iterator last = l.end();
        for(std::vector<T>::const_iterator it = l.begin(); it!=last; ++it)
            s << *it;

        return s;
    }

template<class T> inline Serializer& operator>>(Serializer& s, std::vector<T>& l)
    {
        Uint32 size;
        s >> size;
	l.resize(size);
	std::vector<T>::iterator last = l.end();
	for(std::vector<T>::iterator it=l.begin(); it!=last; ++it)
	    s >> *it;

        return s;
    }
#else
inline Serializer& operator<<(Serializer& s, std::vector<Uint8> const& l)
  {
  s << (Uint32)l.size();
  std::vector<Uint8>::const_iterator last = l.end();
  for(std::vector<Uint8>::const_iterator it = l.begin(); it!=last; ++it)
    {
    s << (Uint8)*it;
    }

  return s;
  }

inline Serializer& operator>>(Serializer& s, std::vector<Uint8>& l)
  {
  Uint32 size;
  s >> size;
  l.clear();
  while(size--)
    {
    Uint8 tmp;
    s >> tmp;
    l.push_back(tmp);
    }

  return s;
  }
#endif

#endif
