/* $Id: Connection.h,v 1.16 2002/04/07 07:51:17 doomster Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __ARIANNE_CONNECTION_H
#define __ARIANNE_CONNECTION_H

#include "DataTypes/AList.h"
#include "Types.h"
#include "Concurrent/Mutex.h"
#include "Concurrent/Templock.h"

#include "0_Network/GenericMessage.h"

#include "SDL_net.h"
#include <string>

/** This an abstract network connection between two peers. */
class Connection
  {
  private:
  /** Buffers to prevent uncontrolled behaviour */
  std::vector<Uint8> DataToSend;
  Mutex BlockSend;
  std::vector<Uint8> DataReceived;
  Mutex BlockReceived;
   
  protected:
  /** TPC/IP v4 */
  std::string Host;
  Uint16 Port;

  /** Actual number of bytes send */
  Uint32 DataSent;
  /** Actual number of bytes recieve */
  Uint32 DataRecv;

  //TODO: Move the interface to std::vector
  virtual Sint32 Send(const Uint8 *Data,Uint32 size,Uint32 n_elems)=0;
  virtual Sint32 Recv(Uint8 *Data,Uint32 size,Uint32 n_elems)=0;
  virtual Sint32 BlockingSend(const Uint8 *Data,Uint32 size,Uint32 n_elems)=0;
  virtual Sint32 BlockingRecv(Uint8 *Data,Uint32 size,Uint32 n_elems)=0;

  Connection(const Connection& b);
  Connection& operator=(Connection& b);

  public:
  friend class ConnectionSet;

  Connection();
  Connection(const char *Host, Uint16 Port);
  virtual ~Connection();

  /** This set the Host where the connection will be done */
  void SetHost(const char *Host, Uint16 Port);

  /** This Open the connection to the Host. */
  virtual void Open()=0;
  /** This close the connection to the Host. */
  virtual void Close()=0;
  /** This accept a new connection to the Host. */
  virtual Connection *Accept()=0;

  /** This sent a generic message through the connection */
  Connection& operator <<(GenericMessage &gm);
  /** This recieve a generic message through the connection */
  Connection& operator >>(GenericMessage &gm);

  /** This method returns data from the connection */
  void GetStatistics(Uint32 *DataSent,Uint32 *DataRecv);
  /** This method set data of the connection */
  void SetStatistics(Uint32 DataSent,Uint32 DataRecv);

  virtual void SendPrompt();
  virtual bool RecvPrompt();

  virtual void SendBuffer();
  virtual void ReceiveBuffer();
  
  /** This returns true if there is data waiting on the connection */
  virtual bool isData()=0;
  /** This returns true if there is a message ready to be taken */
  bool isMessage();

  class ConnectionException: public Exception
  {
      protected:
          ConnectionException(): Exception() {}
  };

  class ConnectionOpenException: public ConnectionException
  {
      public:
          ConnectionOpenException(std::string const &where = "UNKNOWN"): ConnectionException()
          {
              message = where + " - Error opening connection";
          }
  };

  class ConnectionAcceptException: public ConnectionException
  {
      public:
          ConnectionAcceptException(std::string const &where = "UNKNOWN"): ConnectionException()
          {
              message = where + " - Error accepting connection: Could not accept connection.";
          }
  };

  class ConnectionReadWriteException: public ConnectionException
  {
      protected:
          ConnectionReadWriteException():ConnectionException() {}
  };

  class ConnectionReadException: public ConnectionReadWriteException
  {
      public:
          ConnectionReadException(std::string const &where = "UNKNOWN"): ConnectionReadWriteException()
          {
              message = where + " - Error reading from connection.";
          }
  };

  class ConnectionWriteException: public ConnectionReadWriteException
  {
      public:
          ConnectionWriteException(std::string const &where = "UNKNOWN"): ConnectionReadWriteException()
          {
              message = where + " - Error writting to connection.";
          }
  };

  class ConnectionNoMessageException: public ConnectionException
  {
       public:
	   ConnectionNoMessageException(std::string const &where = "UNKNOWN"): ConnectionException()
	   {
	       message = where + " - No Message waiting in the buffer.";
	   }
  };


  };


/** This is an abstraction of the select call of Unix */
class ConnectionSet
  {
  protected:
  SDLNet_SocketSet socketset;  
    						 
  public:
  ConnectionSet(Uint32 NumConnections);
  virtual ~ConnectionSet();

  /** This add a connection to the Set */
  virtual void AddConnection(Connection *conn)=0;
  /** This remove a connection to the Set */
  virtual void RemoveConnection(Connection *conn)=0;

  /** This blocks the caller here until a connection arrive or timeout. */
  void Wait(Uint32 time);
  /** This returns true if there is data waiting on the connection */
  virtual bool isData(Connection *conn)=0;
  /** This returns true if there is a message ready to be taken */
  bool isMessage(Connection *conn);

  class ConnectionSetException: public Exception
  {
      protected:
          ConnectionSetException(): Exception() {}
  };

  class ConnectionSetOpenException: public ConnectionSetException
  {
      public:
	  ConnectionSetOpenException(std::string const &where = "UNKNOWN"): ConnectionSetException()
	  {
	      message = where + " - Error creating connection set.";
	  }
  };

  class ConnectionSetAddException: public ConnectionSetException
  {
      public:
	  ConnectionSetAddException(std::string const &where = "UNKNOWN"): ConnectionSetException()
	  {
	      message = where + " - Error adding connection to connection set.";
	  }
  };

  class ConnectionSetRemoveException: public ConnectionSetException
  {
      public:
	  ConnectionSetRemoveException(std::string const &where = "UNKNOWN"): ConnectionSetException()
	  {
	      message = where + " - Error removing connection to connection set.";
	  }
  };

  class ConnectionSetSelectException: public ConnectionSetException
  {
      public:
	  ConnectionSetSelectException(std::string const &where = "UNKNOWN"): ConnectionSetException()
	  {
	      message = where + " - Error waiting on connection set.";
	  }
  };

  };

#endif
