/* $Id: libArianneclient.h,v 1.10 2002/04/30 15:22:46 arianne_rpg Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef ARIANNE_Client
#define ARIANNE_Client

#include "4_Object/Action.h"
#include "4_Object/Object.h"
#include "0_Network/NetworkClientManager.h"
#include "1_Message/MessageClientManager.h"
#include "Crypto/Crypto.h"


/** Class that encapsulate all the client calls to communicate with server. */
class ArianneClient
  {
  // not copyable
  ArianneClient(ArianneClient const&);
  // not assignable
  ArianneClient& operator=(ArianneClient const&);

  public:
  /** Structure for sending information about the client to the server.
      Just for statistical purposes. */
  struct Info
    {
    std::string CPU;
    std::string OS;
    };

  /** Structure to send information about client statistics. */
  struct Stats
    {
    std::string ClientID;
    std::string ClientVersion;
    std::string Frames;
    };

  /** type to choose from various login profiles */
  enum logintype {password, rsakey};

  /** Structure to host all the login info. */ 
  struct LoginInfo
    {
    std::string Host;
    Uint16 Port;
    std::string Name;
    logintype Type;
    std::string Password;
    std::string RSAKeyLocation;
    };

  private:
  /** The Network Manager */
  NetworkClientManager NetMan;
  /** The Message Manager */
  MessageClientManager MesMan;

  /** An array of list for buffering perceptions (0) and actions(1). */
  std::list<Message*> MessageBuffer[2];
  /** true if Login Stage is complete. */
  bool LoginStageComplete;
  /** true if Logout Stage is complete. */
  bool LogoutStageComplete;

  LoginInfo login;

  void AddController(Object::ID const& id);

  /** List of object::ID that the Client controls. */
  std::list<Object::ID> Controller;

#ifdef HAVE_CRYPTO
  //TODO: Do this need class scope?
  std::vector<unsigned char> h_w1;
  std::vector<unsigned char> hash;
  std::vector<unsigned char> w1;
  std::vector<unsigned char> w2;
  std::vector<unsigned char> m;
  std::vector<unsigned char> signature;
  RSAKey *key;
  RAND r;
#endif

  public:
  ArianneClient();
  ~ArianneClient();

  /** This method makes the client contact the server and login according to
  the given LoginInfo. It throws various kinds of exceptions to signal failure.
  */
  void Login(LoginInfo const& login);
  /** This method finishes the actual session, flushes the message-queues and
  drops the connection to the server. It throws various kinds of exceptions to
  signal failure.
  */
  void Logout(Stats const& stats);

  /** Returns true if client is correctly logged in. */
  bool isLogin();

  /** Wait until a message arrives, blocking if there is none available. */
  void WaitMessage();
  /** Proccess all the available messages without blocking. */
  void PumpMessages();

  /** Send the system info inmediately. */
  void SendSystemInfo(Info const& info);
 
  /** Send a list of actions inmediately. */
  void SendAction(std::list<Action> const& actions);

  /** If available, it stores on the list the data about objects modified and deleted since
      last percetion message, otherwise it throws an InvalidData exception. */
  void RecievePerception(Uint32 &timestamp, std::list<Object> &modified, std::list<Object> &deleted);
  /** Return true if there is a perception available. */
  bool isAvailablePerception();
  /** If available, it stores on the list the data about actions that the client should be
      aware of, so that client can represent them on GFX, SFX or AI. Otherwise it throws an
      InvalidData exception. */
  void RecieveAction(std::list<std::list<Action> > &actions);
  /** Returns true if there is an action available. */
  bool isAvailableAction();

  // TODO: Actually Arianne support only one Object per client.
  /** Return the list with IDs of the Objects that this client controls. */
  std::list<Object::ID> const& GetController();


  class HostNotFound: public Exception
    {
    public:
    HostNotFound(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Host not found.";
      }
    };

  class BadLogin: public Exception
    {
    public:
    BadLogin(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Bad Login.";
      }
    };

  class ConnectionLost: public Exception
    {
    public:
    ConnectionLost(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Connection to server has been lost.";
      }
    };

  class NotLogged: public Exception
    {
    public:
    NotLogged(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - You are not logged into server.";
      }
    };

  class InvalidData: public Exception
    {
    public:
    InvalidData(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Invalid data found.";
      }
    };
  };


#endif
