/* $Id: NetworkServerManager.cpp,v 1.32 2002/04/26 18:03:39 doomster Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#include "Log/Logger.h"
#include "Concurrent/Templock.h"

#include "0_Network/NetworkServerManager.h"
#include "2_Command/Command.h"
#include "2_Command/CommandManager.h"
#include "3_Manager/ManagerCoordinator.h"

NetworkServerManager::NetworkServerManager ():
  LoopThread("NetworkManager"),
  NetworkManager(),
  conns()
  {
  LastTimeStatCommand = 0;
  glog_Layer0.Debug ("Creating NetworkManager on Server -DONE- ");
  }

NetworkServerManager::~NetworkServerManager ()
  {
  }


void NetworkServerManager::Open (char const* Host, Uint16 Port)
  {
  conns.Open(Host,Port);

  Continue = true;
  }

void NetworkServerManager::Close ()
  { 
  conns.Close();
  }


void NetworkServerManager::Add (GenericMessage * GenMes)
  {
  TempLock templock (mutex);

  try
    {
    PlayerConnection& pc = conns.find(GenMes->ID);
    try
      {
      /** send the message and update the timeout-timer */
      pc.LastMessageTime = SDL_GetTicks ();
      *pc.conn << *GenMes;

      /* update the statistics about the amount of data Sent/Recv */
      Uint32 sent, recv;
      pc.conn->GetStatistics (&sent, &recv);
      pc.conn->SetStatistics (0, 0);

      StatsL0.BytesSentSesion += sent;
      StatsL0.BytesRecvSesion += recv;

      /* Take special actions if the message was a logout-message */
      if (GenMes->Type == ARIANNE_LOGOUT)
        {
        glog_Layer0.Debug ("Sent LOGOUT: Closing connection");
        ClosePlayerConnection (pc);
        }

      delete GenMes;
      }
    catch(Connection::ConnectionException &e)
      {
      ClosePlayerConnection (pc);
      delete GenMes;
      GenMes=0;

      throw NetManUnusedConException (__FILE_AND_LINE__);
      }
    }
  catch(Connections::ConnectionsException &e)
    {
    throw NetManUnusedConException (__FILE_AND_LINE__);
    }
  }

GenericMessage *NetworkServerManager::Get ()
  {
  return incoming.Get ();
  }

bool NetworkServerManager::isAvailable () const
  {
  return incoming.isAvailable ();
  }


void NetworkServerManager::WaitData ()
  {
  conns.WaitData();
  }

void NetworkServerManager::AcceptPlayerConnection ()
  {
  if(conns.isData())
    {
    Connections::ID id=conns.AddPlayer();
    if(id==ERROR_CONNECTION_ID)
      {
      return;
      }

    PlayerConnection& pc = conns.find(id);

    /* We have to send the prompt */
    pc.conn->SendPrompt();

    /* We have to notify for new user */
    NewConnectionCommand *newComm = new NewConnectionCommand;
    newComm->cID = id;
    newComm->ipAddr = ((TCPConnection *)pc.conn)->GetIPAddress();
    CommManager->AddCommand (newComm);
    ++StatsL0.NumberOfConnections;
    }
  }

void NetworkServerManager::ReadPlayerConnection (PlayerConnection &pc)
  {
  GenericMessage *gm = new GenericMessage;
  *pc.conn >> *gm;

  gm->ID = pc.id;
  pc.LastMessageTime = gm->TimeToken = SDL_GetTicks ();

  incoming.Add (gm);
  }

void NetworkServerManager::ClosePlayerConnection (PlayerConnection &pc)
  {
  try
    {
    pc.conn->SendBuffer();
    }
  catch(Connection::ConnectionReadWriteException & e)
    {
    }
  conns.DelPlayer(pc.id);
  }

void NetworkServerManager::SendPoisonToUpperLayer ()
  {
  GenericMessage *gm = new GenericMessage;
  gm->Type = ARIANNE_POISON;
  gm->Flags = 0;
  gm->ID = ERROR_CONNECTION_ID;

  incoming.Add (gm);
  }

void NetworkServerManager::SentStatL0Command (bool Mandatory)
  {
  if (((SDL_GetTicks () - LastTimeStatCommand) >TIME_BETWEEN_STATS_COMMANDS) || (Mandatory))
    {
    LastTimeStatCommand = SDL_GetTicks ();
    /* We send the command to the manager. */
    StatL0Command *sL0Comm = new StatL0Command;
    sL0Comm->tmpInfL0 = StatsL0;
    StatsL0.Reset ();
    CommManager->AddCommand (sL0Comm);
    } 
  }

void NetworkServerManager::loop ()
  {
  WaitData ();

  SentStatL0Command ();

  if(!Continue)
    {
    SentStatL0Command (true);

    /* Finish this layer and notifying about it to the upper one */
    //SendPoisonToUpperLayer ();

    Stop ();
    return;
    }

  TempLock templock (mutex);
  
  AcceptPlayerConnection ();


  for (Connections::iterator it=conns.begin();it!=conns.end();++it)
    {
    try
      {      
      (*it).conn->SendBuffer();
      (*it).conn->ReceiveBuffer();

      if(conns.isMessage((*it).id))
        {
        try
          {
          ReadPlayerConnection ((*it));

          /* We increment the amount of data Sent/Recv */
          Uint32 bSent, bRecv;
          (*it).conn->GetStatistics (&bSent, &bRecv);
          (*it).conn->SetStatistics (0, 0);

          StatsL0.BytesSentSesion += bSent;
          StatsL0.BytesRecvSesion += bRecv;
          }
        catch(Connection::ConnectionNoMessageException &e)
          {
          glog_Layer0.Debug (__FILE_AND_LINE__, e.what (),Log::error);
          }
        }
      }
    catch (Connection::ConnectionReadWriteException & e)
      {
      glog_Layer0.Debug (__FILE_AND_LINE__, e.what (),Log::error);

      /* We notify about the event to the Admins */
      CloseConnectionCommand *ccComm = new CloseConnectionCommand;
      ccComm->cID = (*it).id;
      CommManager->AddCommand (ccComm);

      /* There was an error reading or writting on Connection, so we kill it */
      ClosePlayerConnection (*it);

      /* If we delete any player we should reset the iterator. */
      it=conns.begin();

      /* We increment the number of Networks errors */
      ++StatsL0.NetworkErrors;
      }
    }
  }
