/* $Id: PlayerConnectionManager.cpp,v 1.32 2002/04/23 09:46:39 arianne_rpg Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <string>
#include <sstream>

#include <stdio.h>

#include "Extern.h"
#include "Conversion/Conversion.h"
#include "Crypto/Crypto.h"
#include "Log/Logger.h"

#include "3_Manager/PlayerConnectionManager.h"

PlayerInfo::PlayerInfo(WorldManager *worldMan):
  ConnectionID(ERROR_CONNECTION_ID),
  PlayerID(Object::Invalid),
  loginProcedureInfo(worldMan)
  {
  loginProcedureInfo.StateOfID=invalid;
  loginProcedureInfo.IPaddress="";
  perceptionInfo.LastPerception.clear();
  }

PlayerInfo::~PlayerInfo()
  {
  }

#ifdef HAVE_CRYPTO
const std::vector<Uint8> PlayerInfo::LoginProcedureInfo::GetHOfw1()
  {
  RAND r;
  std::vector<Uint8> result;

  r.GetRandomBytes(cryptoLoginExtraInfo.w1, 20);
  SHA1::GetHash(cryptoLoginExtraInfo.w1, result);
  StateOfID = wait_w2;

  return result;
  }

bool PlayerInfo::LoginProcedureInfo::Setw2(std::vector<Uint8> const &w2)
  {
  if(w2.size() != 20)
    {
    return false;
    }

  cryptoLoginExtraInfo.w2 = w2;
  cryptoLoginExtraInfo.m.resize(20);
 
  for(int i=0;i<20;i++)
    {
    /* We made the XOR between W1 and W2 */
    cryptoLoginExtraInfo.m[i]=cryptoLoginExtraInfo.w1[i]^cryptoLoginExtraInfo.w2[i];
    }

  StateOfID = wait_signature;
  return true;
  }

std::vector<Uint8> const &PlayerInfo::LoginProcedureInfo::Getm()
  {
  return cryptoLoginExtraInfo.m;
  }

bool PlayerInfo::LoginProcedureInfo::VerifyRSA(std::vector<Uint8> const &signature)
{
  RSAKey *r = LoginInformation.getKey();
  if(!r)
  {
    return false;
  }

  bool verifyOK = r->VerifySignature(Getm(), signature);
  if (!verifyOK)
  {
    LoginInformation.increaseFailures();
    try
    {
      GlobalDb->updateLogin(LoginInformation);
    }
    catch (Exception &e)
    {
      glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }
    return false;
  }

  return true;
}
#endif

PlayerInfo::LoginProcedureInfo::LoginProcedureInfo(WorldManager *worldMan)
{
  this->worldMan=worldMan;
}

/** We begin the login procedure, because we received a NLogin */
bool PlayerInfo::LoginProcedureInfo::BeginLogin(std::string const &login)
{
  LoginInformation.setLogin(login);
  try
  {
    GlobalDb->getLogin(login,LoginInformation);
  }
  catch (Exception &e)
  {
    glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
    throw;
  }
  if(!LoginInformation.isValid())
  {
    glog_Layer3.Debug("Incorrect login for '" + login + "'", Log::error);
    return false;
  }

  StateOfID = waitPassword;

  #ifdef HAVE_CRYPTO
  if(LoginInformation.getLoginType() == LoginInfo::RSA)
  {
    StateOfID = create_w1;
  }
  #endif

  Object::ID oid = LoginInformation.getObjectID();
  if (oid == Object::Invalid)
  {
    /* We have a valid id to the Object and create a new object for it */
    glog_Layer3.Debug("First Player's Login: Creating a character");
    oid = worldMan->GetValidID();
    LoginInformation.setObjectID(oid);

    try
    {
      GlobalDb->setLogin(LoginInformation);
    }
    catch (Exception &e)
    {
      glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }

    // TODO: The Type of character along with other params should be get from a table. */
    worldMan->CreateRandomPlayer("human",LoginInformation.getLogin(),oid);
  }
  else
  {
    try
    {
      worldMan->Get(oid);
    }
    catch(Object::ObjectNotFoundException &e)
    {
      /* Object is on Database but is not in world. */
      glog_Layer3.Debug("Database-World Mismatch: Creating a character");

      // TODO: The Type of character along with other params should be get from a table. */
      worldMan->CreateRandomPlayer("human",LoginInformation.getLogin(),oid);
    }
  }
 
  return true;
}

/** We verify the password. */
bool PlayerInfo::LoginProcedureInfo::VerifyLogin(std::string const &pass)
{
  #ifdef HAVE_CRYPTO
  bool verifyOK = SHA1::verifyPW(pass, LoginInformation.getHash());
  #else
  bool verifyOK = (pass == LoginInformation.getHash());
  #endif

  if (!verifyOK)
  {
    LoginInformation.increaseFailures();
    try
    {
      GlobalDb->updateLogin(LoginInformation);
    }
    catch (Exception &e)
    {
      glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }
  }

  return verifyOK;
}

/** We finish the login stage by setting Object::ID and updating a few fields. */
bool PlayerInfo::LoginProcedureInfo::CompleteLogin()
{
  LoginInformation.setFailures(0);
  LoginInformation.setCurrentIP(IPaddress);
  try
  {
    GlobalDb->updateLogin(LoginInformation);
  }
  catch (Exception &e)
  {
    glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
    throw;
  }

  StateOfID=valid;

  return true;
}

std::string PlayerInfo::LoginProcedureInfo::GetLastLogin() const
  {
  std::ostringstream sout;

  sout << "Last login on " << LoginInformation.getLastDateTime()
       << " from " << LoginInformation.getLastIP()
       << " after " << LoginInformation.getFailures() << " failures.";

  return sout.str();
  }


bool PlayerInfo::LoginProcedureInfo::isReady() const
  {
  return (StateOfID == PlayerInfo::valid);
  }

PlayerConnectionManager::PlayerConnectionManager (WorldManager *wm):
  WorldMan(wm)
  {
  }

PlayerConnectionManager::~PlayerConnectionManager ()
  {
  }


void PlayerConnectionManager::Use (PlayerConnectionManager::ID i)
  {
  TempLock tmp(mutex);

  if(find(i)!=end())
    {
    /* This ID is already in use. This is not an error, this is a bug */
    assert(false);
    }

  PlayerInfo player(WorldMan);
  player.ConnectionID=i;
  
  push_back(player);
  }

void PlayerConnectionManager::Release (PlayerConnectionManager::ID i)
  {
  TempLock tmp(mutex);

  iterator it=find(i);
  if(it==end())
    {
    /* This can happen as a result of closing twice a connection...
       Not sure if it is a bug or can really happen. */
    throw ManagerUnUsedConnectionException (__FILE_AND_LINE__);
    }

  erase(it);
  }

PlayerConnectionManager::iterator PlayerConnectionManager::find(PlayerConnectionManager::ID i)
  {
  TempLock tmp(mutex);

  for(iterator it=begin();it!=end();++it)
    {
    if((*it).ConnectionID==i)
      {
      return it;
      }
    } 

  return end();
  }

PlayerConnectionManager::iterator PlayerConnectionManager::find(Object::ID const& i)
  {
  TempLock tmp(mutex);

  for(iterator it=begin();it!=end();++it)
    {
    if((*it).PlayerID==i)
      {
      return it;
      }
    } 

  return end();
  }



#if 0
PlayerInfo::PlayerInfo():
    used(false),
    ConnectionID(MAX_NUMBER_CONNECTIONS),
    PlayerID(Object::Invalid),
    stateOfID(invalid),
    IPaddr(""),
    l(0)
  {
  }

PlayerInfo::~PlayerInfo()
  {
  delete l;
  l = 0;
  }

PlayerConnectionManager::PlayerConnectionManager (WorldManager *wm):
    WorldMan(wm)
  {
  }

PlayerConnectionManager::~PlayerConnectionManager ()
{
}

bool PlayerConnectionManager::BeginLogin(Uint32 i, std::vector<unsigned char> const& vectlogin)
{
  TempLock tmp (mutex);
  std::string login;

  PlayerInfo* pi = pconn+i;

  VectorToString(vectlogin, login);

  pi->l = new LoginInfo(login);
  try
  {
    GlobalDb->getLogin(login, *pi->l);
  }
  catch (Exception &e)
  {
    glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
    throw;
  }

  if(!pi->l->isValid())
  {
    std::cerr << "incorrect login for '" << login << "'\n";
    delete pi->l;
    pi->l=0;
    return false;
  }
  
  pi->stateOfID = PlayerInfo::waitPassword;
#ifdef HAVE_CRYPTO
  if(pi->l->getLoginType() == LoginInfo::RSA)
    pi->stateOfID = PlayerInfo::create_w1;
#endif

  //TODO: Need to be change to adapt to new Brian's code.
  Object::ID oid=pi->l->getObjectID();
  if(oid==Object::Invalid)
  {
    /* We have a valid id to the Object and create a new object for it */
    glog_Layer3.Debug("First Player's Login: Creating a character");
    oid=WorldMan->GetValidID();
    pi->l->setObjectID(oid);
    try
    {
      GlobalDb->setLogin(*pi->l);
    }
    catch (Exception &e)
    {
      glog_Layer3(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }
    //TODO: Need to be randomly choosed. 
    WorldMan->CreateRandomPlayer("human",pi->l->getLogin(),oid);
  }
 
  return true;
}

bool PlayerConnectionManager::VerifyLogin(Uint32 i, std::vector<unsigned char> const &vectpass)
{
  assert(pconn[i].l);
  TempLock tmp (mutex);
  std::string pass;
  VectorToString(vectpass, pass);
  
#ifdef HAVE_CRYPTO
  //  This line was for using MD5 Hash instead.
  //  bool verifyOK= MD5Hash(pass)==pconn[i].l->getHash();
  bool verifyOK = SHA1::verifyPW(pass, pconn[i].l->getHash());
#else
  bool verifyOK = (pass == pconn[i].l->getHash());
#endif

  if (!verifyOK)
  {
    pconn[i].l->increaseFailures();
    try
    {
      GlobalDb->updateLogin(*pconn[i].l);
    }
    catch (Exception &e)
    {
      glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }
  }

  return verifyOK;
}

bool PlayerConnectionManager::isUsed (Uint32 i) const
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  return pconn[i].used;
}

#ifdef HAVE_CRYPTO
const std::vector<unsigned char> PlayerConnectionManager::GetHOfw1(Uint32 cID)
{
  RAND r;
  std::vector<unsigned char> res;
  TempLock tmp (mutex);
  r.GetRandomBytes(pconn[cID].w1, 20);
  SHA1::GetHash(pconn[cID].w1, res);
  pconn[cID].stateOfID = PlayerInfo::wait_w2;
  return res;
}

bool PlayerConnectionManager::Setw2(Uint32 cID, std::vector<unsigned char> const &w2)
{
  TempLock tmp (mutex);
  std::vector<unsigned char>::iterator i,j,k,last;
  if(w2.size() != 20)
    return false;
  pconn[cID].w2 = w2;
  pconn[cID].m.resize(20);
  last = pconn[cID].w2.end();
  for(i = pconn[cID].m.begin(), j=pconn[cID].w1.begin(), k=pconn[cID].w2.begin();k!=last; ++i,++j,++k)
      *i = *j ^ *k;
  pconn[cID].stateOfID = PlayerInfo::wait_signature;
  return true;
}

std::vector<unsigned char> const &PlayerConnectionManager::Getm(Uint32 cID)
{
  TempLock tmp (mutex);
  return pconn[cID].m;
}

bool PlayerConnectionManager::VerifyRSA(Uint32 cID, std::vector<unsigned char> const &signature)
{
  RSAKey *r;
  TempLock tmp (mutex);

  r = pconn[cID].l->getKey();
  if (!r)
    return false;
  bool verifyOK = r->VerifySignature(pconn[cID].m, signature);

  if (!verifyOK)
  {
    pconn[cID].l->increaseFailures();
    try
    {
      GlobalDb->updateLogin(*pconn[cID].l);
    }
    catch (Exception &e)
    {
      glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
      throw;
    }
  }

  return verifyOK;
}
#endif

Uint32 PlayerConnectionManager::BindPlayerToConnection (Uint32 cID)
{
  assert(pconn[cID].l);
  Object::ID pID;

  {
      TempLock tmp (mutex);
      pID = pconn[cID].l->getObjectID();
  }

  return BindPlayerToConnection (cID, pID);
}

Uint32 PlayerConnectionManager::BindPlayerToConnection (Uint32 cID, Object::ID pID)
{
  assert(pconn[cID].l);
  TempLock tmp (mutex);

  Uint32 OldCID;
  int i;

  pconn[cID].l->setFailures(0);
  pconn[cID].l->setCurrentIP(pconn[cID].IPaddr);
  try
  {
    GlobalDb->updateLogin(*pconn[cID].l);
  }
  catch (Exception &e)
  {
    glog_Layer3.Debug(__FILE_AND_LINE__, e.what(), Log::error);
    throw;
  }

  OldCID = MAX_NUMBER_CONNECTIONS;
  for (i = 0; i < MAX_NUMBER_CONNECTIONS; ++i)
    if ((pconn[i].used == true) && (pconn[i].PlayerID == pID))
      OldCID = i;

  if (OldCID != MAX_NUMBER_CONNECTIONS)
  {
    /* The new ID must take the place of the older one. */
    pconn[cID].ConnectionID = cID;
    pconn[cID].PlayerID = pID;
    pconn[cID].stateOfID = PlayerInfo::valid;
    pconn[cID].LastPerception = pconn[OldCID].LastPerception;
  }
  else
  {
    pconn[cID].ConnectionID = cID;
    pconn[cID].PlayerID = pID;
    pconn[cID].stateOfID = PlayerInfo::valid;
    pconn[cID].LastPerception.clear();
  }

  if(pconn[cID].l)
  {
    delete(pconn[cID].l);
    pconn[cID].l = 0;
  }

  return OldCID;
}

void
PlayerConnectionManager::Use (Uint32 i, std::string ip)
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  if (pconn[i].used)
  {
    throw ManagerUsedConnectionException (__FILE_AND_LINE__);
  }

  pconn[i].used = true;
  pconn[i].PlayerID = Object::Invalid;
  pconn[i].stateOfID = PlayerInfo::invalid;
  pconn[i].LastPerception.clear ();
  pconn[i].IPaddr = ip;
}

void
PlayerConnectionManager::Release (Uint32 i)
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  if (!pconn[i].used)
  {
    throw ManagerUnUsedConnectionException (__FILE_AND_LINE__);
  }

  pconn[i].used = false;
  pconn[i].PlayerID = Object::Invalid;
  pconn[i].stateOfID = PlayerInfo::invalid;
  pconn[i].LastPerception.clear ();
  if(pconn[i].l)
  {
    delete pconn[i].l;
    pconn[i].l = 0;
  }
}

Object::ID PlayerConnectionManager::GetPlayerID (Uint32 cID) const
{
  assert (cID < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  return pconn[cID].PlayerID;
}

Uint32 PlayerConnectionManager::GetConnectionID (Object::ID pID) const
{
  TempLock tmp (mutex);

  for (int i = 0; i < MAX_NUMBER_CONNECTIONS; ++i)
  {
    if ((pconn[i].used == true) && (pconn[i].PlayerID == pID))
    {
      return pconn[i].ConnectionID;
    }
  }

  throw ManagerUnUsedConnectionException (__FILE_AND_LINE__);
  return MAX_NUMBER_CONNECTIONS;
}

std::string PlayerConnectionManager::GetLastLogin (Uint32 cID) const
{
  assert(pconn[cID].l);
  TempLock tmp (mutex);

  std::ostringstream sout;

  sout << "Last login on " << pconn[cID].l->getLastDateTime()
       << " from " << pconn[cID].l->getLastIP()
       << " after " << pconn[cID].l->getFailures() << " failures.";

  return sout.str();
}

void
PlayerConnectionManager::Ready (Uint32 i)
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].stateOfID = PlayerInfo::valid;
}

bool PlayerConnectionManager::isReady (Uint32 i) const
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  return (pconn[i].stateOfID == PlayerInfo::valid);
}

PlayerInfo::States PlayerConnectionManager::GetStateOfID(Uint32 i) const
{
  TempLock tmp (mutex);
  return pconn[i].stateOfID;
}

std::list<Object> PlayerConnectionManager::GetLastPerception (Uint32 i) const
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  return pconn[i].LastPerception;
}

void
PlayerConnectionManager::SetLastPerception (Uint32 i,
					    std::list < Object > const& NewPerception)
{
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].LastPerception.clear ();
  pconn[i].LastPerception = NewPerception;
}

void PlayerConnectionManager::SetActionToDeliver(Uint32 i, std::list<Action> const& action)
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].ActionsToDeliver=action;
  }
  
void PlayerConnectionManager::ClearActionToDeliver(Uint32 i)
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].ActionsToDeliver.clear();
  }

std::list<Action> PlayerConnectionManager::GetActionToDeliver(Uint32 i) const
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  std::list<Action> action=pconn[i].ActionsToDeliver;

  return action;
  }

void PlayerConnectionManager::SetActionsToSend(Uint32 i, std::list<std::list<Action> > const& actions)
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].ActionsToSend=actions;
  }

void PlayerConnectionManager::AddActionsToSend(Uint32 i, std::list<Action> const& action)
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].ActionsToSend.push_back(action);
  }


void PlayerConnectionManager::ClearActionsToSend(Uint32 i)
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  pconn[i].ActionsToSend.clear();
  }

std::list<std::list<Action> > PlayerConnectionManager::GetActionsToSend(Uint32 i) const
  {
  assert (i < MAX_NUMBER_CONNECTIONS);
  TempLock tmp (mutex);

  return pconn[i].ActionsToSend;
  }

#endif
