/* $Id: PlayerConnectionManager.h,v 1.26 2002/04/09 12:06:24 arianne_rpg Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef ARIANNE_PLAYERCONNECTIONMANAGER
#define ARIANNE_PLAYERCONNECTIONMANAGER

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Types.h"
#include "Define.h"
#include "Manager/Manager.h"
#include "Concurrent/Mutex.h"
#include "Concurrent/Templock.h"

#include "3_Manager/LoginInfo.h"
#include "4_Object/Object.h"
#include "4_Object/Action.h"
#include "5_Rp/WorldManager.h"

/** This contains all the information about players, that isn't related to the Object. */ 
struct PlayerInfo
  {
  PlayerInfo(WorldManager *worldMan);
  ~PlayerInfo();

  /** The connection identificator. */
  Uint32 ConnectionID;
  /** The player identification */
  Object::ID PlayerID;

  /** Is the player ready for start the game? == Where are we in the state of the Identification. */
  enum States
    {
    invalid,
    waitPassword,
    #ifdef HAVE_CRYPTO
    create_w1,
    wait_w2,
    wait_signature,
    #endif
    valid
    };

  struct LoginProcedureInfo
    {
    /** Internal state of the PlayerConnection */
    States StateOfID;

    /** The IP address of the client machine */
    std::string IPaddress;

    /** Login information for the player, we can delete that as soon as he is identified. */
    LoginInfo LoginInformation;

    #ifdef HAVE_CRYPTO
    struct CryptoLoginExtraInfo
      {
      /** The differents vector we'll need to remember to grant access to a player using RSA */
      std::vector<unsigned char> w1;
      std::vector<unsigned char> w2;
      std::vector<unsigned char> m;
      std::vector<unsigned char> signature;
      };

    /** The procedures to handle RSA login */
    const std::vector<Uint8> GetHOfw1();
    bool Setw2(std::vector<Uint8> const &w2);
    std::vector<Uint8> const &Getm();
    bool VerifyRSA(std::vector<Uint8> const &signature);
    
    CryptoLoginExtraInfo cryptoLoginExtraInfo;
    #endif

    WorldManager *worldMan;

    LoginProcedureInfo(WorldManager *worldMan);

    /** We begin the login procedure, because we received a NLogin */
    bool BeginLogin(std::string const &login);
    /** We verify the password. */
    bool VerifyLogin(std::string const &vectpass);
    /** We finish the login stage by setting Object::ID and updating a few fields. */
    bool CompleteLogin();

    /** Is the login stage completed for this player? */
    bool isReady() const;
    /** Returns the string with the info about the last login. */
    std::string GetLastLogin() const;
    };

  LoginProcedureInfo loginProcedureInfo;

  struct PerceptionInfo
    {
    /** The last perception sent to the object */
    std::list<Object> LastPerception;
    /** The set of actions that need to be send to all client from this client. */
    std::list<Action> ActionsToDeliver;
    /** The set of actions that need to be send to this client about the rest of clients. */
    std::list<std::list<Action> > ActionsToSend;
    };

  PerceptionInfo perceptionInfo;
  };

/** This class encapsulates a Player Info for each player but it is just a 
    container class, it has no active behaviour. */
class PlayerConnectionManager: public std::list<PlayerInfo>
  {
  Mutex mutex;
  WorldManager *WorldMan;

  public:
  typedef Uint32 ID;
  PlayerConnectionManager(WorldManager *WorldMan);
  ~PlayerConnectionManager();

  /** Player start to use entry i. */
  void Use(PlayerConnectionManager::ID i);
  /** Player release entry i.*/
  void Release(PlayerConnectionManager::ID i);

  PlayerConnectionManager::iterator find(PlayerConnectionManager::ID i);
  PlayerConnectionManager::iterator find(Object::ID const& i);

  class ManagerUsedConnectionException: public Exception
    {
    public:
    ManagerUsedConnectionException(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Error: Trying to acquire an used connection.";
      }
    };

  class ManagerUnUsedConnectionException: public Exception
    {
    public:
    ManagerUnUsedConnectionException(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Error: Trying to release an unused connection.";
      }
    };
  };



#if 0
/** This contains all the information about players, that isn't related to the Object. */ 
struct PlayerInfo
  {
  PlayerInfo();
  ~PlayerInfo();

  /** Is this slot being used? */
  bool used;
  /** The connection identificator. */
  Uint32 ConnectionID;
  /** The player identification */
  Object::ID PlayerID;
  /** The last perception sent to the object */
  std::list<Object> LastPerception;
  /** The set of actions that need to be send to client. */
  std::list<Action> ActionsToDeliver;
  std::list<std::list<Action> > ActionsToSend;

  /** Is the player ready for start the game? == Where are we in the state of the Identification. */
  enum States{ invalid,
    waitPassword,
#ifdef HAVE_CRYPTO
    create_w1,
    wait_w2,
    wait_signature,
#endif
    valid };
 
  /** Internal state of the PlayerConnection */
  States stateOfID;

  /** The IP address of the client machine */
  std::string IPaddr;

  /** Login information for the player, we can delete that as soon as he
   is identified. */
  LoginInfo *l;

#ifdef HAVE_CRYPTO
  /** The differents vector we'll need to remember to grant access to a
   player using RSA */
  std::vector<unsigned char> w1;
  std::vector<unsigned char> w2;
  std::vector<unsigned char> m;
  std::vector<unsigned char> signature;
#endif
  };  // struct Playerinfo

/** This class encapsulates a Player Info for each player but it is just a 
    container class, it has no active behaviour. */
class PlayerConnectionManager: public Manager
  {
  Mutex mutex;
  //TODO: It should be converted in to a List.
  /** The array containing the Player info entries. */
  PlayerInfo pconn[MAX_NUMBER_CONNECTIONS];

  WorldManager *WorldMan;

  public:
  PlayerConnectionManager(WorldManager *WorldMan);
  ~PlayerConnectionManager();

  /** Is entry i being used? */
  bool isUsed(Uint32 i) const;
  /** Associate a player with connection cIS to id pID */
  Uint32 BindPlayerToConnection(Uint32 cID,Object::ID pID);
  /** Associate a player with connection cIS to id in LoginInfo */
  Uint32 BindPlayerToConnection(Uint32 cID);
  /** Player start to use entry i. */
  void Use(Uint32 i, std::string ip);
  /** Player release entry i.*/
  void Release(Uint32 i);

  /** Get the player id using its connection id */
  Object::ID GetPlayerID(Uint32 cID) const;
  /** Getthe connection id using the player id. */
  Uint32 GetConnectionID(Object::ID pID) const;
  /** Get info on the last login by this player, must be called before
      BindPlayerToConnection() ! */
  std::string GetLastLogin(Uint32 cID) const;

  /** Set the entry i as ready. */
  void Ready(Uint32 i);
  /** is the entry ready? */
  bool isReady(Uint32 i) const;
  /** Get the current state of identification */
  PlayerInfo::States GetStateOfID(Uint32 i) const;

  /** We begin the login procedure, because we received a NLogin */
  bool BeginLogin(Uint32 i, std::vector<unsigned char> const &vectlogin);
  /** We verify the password. */
  bool VerifyLogin(Uint32 i, std::vector<unsigned char> const &vectpass);
#ifdef HAVE_CRYPTO
  /** The procedures to handle RSA login */
  const std::vector<unsigned char> GetHOfw1(Uint32 cID);
  bool Setw2(Uint32 cID, std::vector<unsigned char> const &w2);
  std::vector<unsigned char> const &Getm(Uint32 cID);
  bool VerifyRSA(Uint32 i, std::vector<unsigned char> const &signature);
#endif

  /** Return a pointer to the LastPerception */
  std::list<Object> GetLastPerception(Uint32 i) const;
  /** Set the last perception for entry i */
  void SetLastPerception(Uint32 i,std::list<Object> const& NewPerception);

  /** This set the actions that the object need to notify to their surround. */
  void SetActionToDeliver(Uint32 i, std::list<Action> const& action);

  void ClearActionToDeliver(Uint32 i);
  /** This returns the actions that the object need to notify to their surround. */
  std::list<Action> GetActionToDeliver(Uint32 i) const;

  /** Add an action that need to be send to client */
  void SetActionsToSend(Uint32 i, std::list<std::list<Action> > const& action);
  void AddActionsToSend(Uint32 i, std::list<Action> const& action);
  void ClearActionsToSend(Uint32 i);
  /** Get the list of actions. */
  std::list<std::list<Action> > GetActionsToSend(Uint32 i) const;



  class ManagerUsedConnectionException: public Exception
    {
    public:
    ManagerUsedConnectionException(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Error: Trying to acquire an used connection.";
      }
    };

  class ManagerUnUsedConnectionException: public Exception
    {
    public:
    ManagerUnUsedConnectionException(std::string const& where = "UNKNOWN"): Exception()
      {
      message = where + " - Error: Trying to release an unused connection.";
      }
    };
  };

#endif

#endif
