/* $Id: RPManager.cpp,v 1.27 2002/04/28 07:48:19 guidoj Exp $ */
/***************************************************************************
    copyright            : (C) 2000 by Miguel Angel Blanch Lardin
    email                : x5101920@fedro.ugr.es
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "RPManager.h"
#include "Log/Logger.h"
#include "Concurrent/Templock.h"
#include "Concurrent/Mutex.h"

#include "SDL.h"


RPManager::RPManager(WorldManager *worldMan):
  Manager(),
  LoopThread("RPManager"),
  m_scheduler_turn(0),
  m_turns(0),
  m_turntime(DEFAULT_TURN_TIME)
  {
  glog_Layer3.Debug ("Creating RPManager on Server");
  this->worldMan=worldMan;
  glog_Layer3.Debug ("Creating RPManager on Server -DONE-");
  }

RPManager::~RPManager()
  {
  }

void RPManager::Add(std::list<Action> const& action)
  {
  /* We suppose the actions have been verified on the ActionManager
     for things as correct user-action source relation. */
  TempLock tmp (mutex);

  // If there is nothing in the action, then do not add it
  if(action.empty())
    {
    return;
    }

  Object::ID ActionSource=action.front().GetSource();

  // Go through every action in the scheduler and if there are any actions
  // from the object that just added this action, replace them
  for(std::list<Entry>::iterator it=m_scheduler[(m_scheduler_turn+1)%2].begin();it!=m_scheduler[(m_scheduler_turn+1)%2].end();++it)
    {
    if((*it).first==ActionSource)
      {
      /** Object has already actions to run... we replace them. */
      (*it).second=action;
      return;
      }
    }

  m_scheduler[(m_scheduler_turn+1)%2].push_back(Entry(ActionSource,action));
  }

std::list<Action> RPManager::Get(Object::ID const& id)
  {
  TempLock tmp (mutex);

  for(std::list<Entry>::iterator it=m_outGoingActions.begin();it!=m_outGoingActions.end();++it)
    {
    if((*it).first==id)
      {
      std::list<Action> tmp=(*it).second;
      m_outGoingActions.erase(it);
      return tmp;
      }
    }

  // TODO: Creating an empty list may be a waste of resource, we may want to consider 
  // returning a NULL pointer... ??
  return std::list<Action>();
  }

std::list<Object> RPManager::GetPerception(Object const& object) const
  {
  /* We should fill this value with RP script one */
  Uint32 range=1000;

  return worldMan->isInRange(object.GetAttrib("position"),range);
  }

std::list<Object> RPManager::GetPerception(Object::ID const& pID) const
  {
  return GetPerception(worldMan->Get(pID));
  }

Uint32 RPManager::SetTurnTime(Uint32 turntime)
{
    if (turntime < MINIMUM_TURN_TIME) turntime = MINIMUM_TURN_TIME;
    if (turntime > MAXIMUM_TURN_TIME) turntime = MAXIMUM_TURN_TIME;
    m_turntime=turntime;
    return m_turntime;
}

void RPManager::loop()
  {
  /* Wait for next turn */

//  #ifndef _TEST_RP_ENGINE
  SDL_Delay(GetTurnTime());
//  #endif

  glog_Layer5.Debug("RP Turn: " + IntToString(m_turns));

  m_turns++;

  /* We run the pending actions. */
  for(std::list<Entry>::iterator it=m_scheduler[m_scheduler_turn].begin();it!=m_scheduler[m_scheduler_turn].end();++it)
    {
    /* It will return false when action is done. */
    ActionResult result=Execute((*it).second.front());

    TempLock tmp (mutex);
    // If the action has been completed.
    if(result==complete_success || result==complete_fail)
      {
      bool found=false;
      // Send Message to client 
      for(std::list<Entry>::iterator it_oga=m_outGoingActions.begin();it_oga!=m_outGoingActions.end();++it_oga)
        {
        // If the first action in the scheduler has the same source as the reply
        // give it the proper status (suucces or failure) and add it to the 
        // outGoingActions queue.
        Object::ID id=(*it).second.front().GetAttrib("source");
        if((*it_oga).first==id)
          {
          Action replyAction=(*it).second.front();
          if(result==complete_success) 
            {
            replyAction.PutAttrib("status","success");
            }
          else
            {
            replyAction.PutAttrib("status","fail");
            }

          (*it_oga).second.push_back(replyAction);
          found=true;
          break;
          } // if((*it_oga).first==id)
        } // for(std::list<Entry>::iterator it_oga=m_outGoingActions.
      
      if(!found)
        {
        Object::ID id=(*it).second.front().GetAttrib("source");
        Action replyAction=(*it).second.front();
        if(result==complete_success) 
          {
          replyAction.PutAttrib("status","success");
          }
        else
          {
          replyAction.PutAttrib("status","fail");
          }
        
        std::list<Action> actions;
        actions.push_back(replyAction);
        m_outGoingActions.push_back(Entry(id,actions));        
        } // if(!found)

      /* If the action is done, we remove it */
      (*it).second.pop_front();  
      } // if(result==complete_success || result==complete_fail)

    if(!(*it).second.empty())
      {
      /* If we find a new action on the next turns that means that the player cancelled the execution of the old ones. */
      bool found=false;

      for(std::list<Entry>::iterator it_nt=m_scheduler[(m_scheduler_turn+1)%2].begin();it_nt!=m_scheduler[(m_scheduler_turn+1)%2].end();++it_nt)
        {
        Object::ID id=(*it).second.front().GetAttrib("source");
        if((*it_nt).first==id)
          {
          found=true;
          break;
          }
        }

      if(!found)
        {
        //TODO: This will mean a lot of copying work, pointer to be considered. 
        m_scheduler[(m_scheduler_turn+1)%2].push_back(*it);
        }
      }     
    }

  /* Next Turn */
  TempLock tmp (mutex);  
  m_scheduler[m_scheduler_turn].clear();
  m_scheduler_turn=(m_scheduler_turn+1)%2;

  if(Commands.isAvailable())
    {
    Command *comm=Commands.Get();

    try
      {
      switch(comm->GetType())
        {
	    case POISON_COMMAND:
  	      /* We finish this manager */
		  delete comm;
		  comm=0;

	      Stop();
		  return;        
		  break;
        default:
          throw ManagerUnCommandException(__FILE_AND_LINE__, comm->GetType());
        }
      }
    catch(Manager::ManagerUnCommandException &e)
      {
      glog_Layer5.Debug(__FILE_AND_LINE__,e.what());

      delete comm;
      comm=0;
      }
    }
  }
