/***************************************************************************
    copyright            : (C) 2002 by Brian Rowe
    email                : briprowe@yahoo.com
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>

// Added include prefixes -waba
#include "PythonLanguage.h"
#include "Log/Logger.h"

using namespace std;

PythonLanguage::PythonLanguage(): pathToExecutable(".")
{
  try 
    {
      Startup();
      // Add the path to the scripts to Python's path
      // g_sharedir is globally defined in InitPath.h
      AddPath(".");
    }
  catch(PyException &e)
    {
      glog_Layer5.Debug(e.what(), Log::fatal_error);
      throw;
    }
}


PythonLanguage::PythonLanguage(const PythonLanguage &rhs)
{
  // TODO
}

PythonLanguage::PythonLanguage(std::string const &path) :
  pathToExecutable(path)
{
 
  // Verify that the path is good
  // If it is not set the path to '.'
  try 
    {
      Startup();
      // Add the path to the scripts to Python's path
      // g_sharedir is globally defined in InitPath.h
      AddPath(path.c_str());
    }
  catch(PyException &e)
    {
      glog_Layer5.Debug(e.what(), Log::fatal_error);
      throw;
    }
}

PythonLanguage::~PythonLanguage()
{
  
  // Stop all running scripts
  // If python is running stop it
  Shutdown(); // This throws an exception, but catching it here
              // will do no good.
  
}

void PythonLanguage::Startup()
{
  if ( !Py_IsInitialized() )
    {
      Py_Initialize();
      if ( PyErr_Occurred() )
	throw PyStartException();
    }
}

void PythonLanguage::Shutdown()
{
  if ( Py_IsInitialized() )
    {
      Py_Finalize();
//    if ( PyErr_Occurred() )
//	  throw PyStopException();
    }
}

void PythonLanguage::AddPath(std::string path)
{
  // assuming that Python is running
  assert(Py_IsInitialized());

  // tell python to load the "sys" module -> it is in charge of the path
  PyRun_SimpleString("import sys");
  if ( PyErr_Occurred() )
    throw PyException();
  
  // create the command to append path to python's path
  std::string command = "sys.path.append(\"" + path + "\")";
  PyRun_SimpleString(const_cast<char*>(command.c_str()));
  if ( PyErr_Occurred() )
    throw PyPathException();
}

PyObject* PythonLanguage::LoadModule(std::string modname)
{
  assert(Py_IsInitialized());

  if ( !modname.length() )
    return NULL;
  
//  PyObject *module = PyModule_New(const_cast<char*>(modname.c_str()));
  PyObject *module = (PyObject*)PyMem_Malloc(sizeof(PyObject));
  module = PyImport_ImportModule(const_cast<char*>(modname.c_str()));
  if ( PyErr_Occurred() )
    {
      glog_Layer5.Debug("failed to load '"+modname+"'\n", Log::warning);
      PyErr_Print();
      Py_XDECREF(module);
      throw PyException();
    }
  return module;
}

PyObject* PythonLanguage::RunFunction(PyObject* module, 
				       std::string funcname,
				       PyObject* args)
{

  if ( !funcname.length() )
    return NULL;

  //PyObject *function; // = (PyObject*)PyMem_Malloc(sizeof(PyObject));
  //PyObject *results;  // = (PyObject*)PyMem_Malloc(sizeof(PyObject));
  
  PyObject *function =PyObject_GetAttrString(module,const_cast<char*>(funcname.c_str()));
  if ( PyErr_Occurred() )
    {
      Py_XDECREF(function);
      // Not used... so I comment it.
      // Py_XDECREF(results);
      Py_XDECREF(args);
      throw PyException();
    }
  PyObject *results = PyObject_CallObject(function, args);
  if ( PyErr_Occurred() )
    {
      Py_XDECREF(function);
      Py_XDECREF(results);
      Py_XDECREF(args);
      throw PyException();
    }
  
  Py_XDECREF(function);
  Py_XDECREF(args);
  return results;

}
/*
PythonLanguage::PyException::PyException(const std::string err)
{
  std::vector<PyObject*> error = GetError();
  errType = error[0];
  errValue = error[1];
  traceback = error[2];

  errorMsg = err;
}

PythonLanguage::PyException::PyException(const char* err)
{
  std::vector<PyObject*> error = GetError();
  errType = error[0];
  errValue = error[1];
  traceback = error[2];
  
  errorMsg = err;
}

PythonLanguage::PyException::PyException()
{
  //  PyObject **errType = (PyObject**)PyMem_Malloc(sizeof(PyObject*));
  //PyObject **errValue = (PyObject**)PyMem_Malloc(sizeof(PyObject*));
  //PyObject **traceback = (PyObject**)PyMem_Malloc(sizeof(PyObject*));
  //PyErr_Fetch(errType, errValue, traceback);

  std::vector<PyObject*> error = GetError();
  errType = error[0];
  errValue = error[1];
  traceback = error[2];

  errorMsg = "";
}

PythonLanguage::PyException::~PyException()
{
  PyErr_Clear();
  Py_XDECREF(errType);
  Py_XDECREF(errValue);
  Py_XDECREF(traceback);
}

std::string PythonLanguage::PyException::what()
{
  std::string retval;

  if ( !errType && errValue )
    retval = "Python Error: " + std::string(PyString_AsString(errValue));
  if ( errType && !errValue )
    retval = "Python Error: " + std::string(PyString_AsString(errType));
  else
    retval = "Python Error: " + std::string(PyString_AsString(errType)) + ": " 
      + std::string(PyString_AsString(errValue));
  
  return(retval);
}

std::vector<PyObject*>& PythonLanguage::PyException::GetError()
{
  PyThreadState *tstate = PyThreadState_Get();

  PyObject *p_type = tstate->curexc_type;
  PyObject *p_val = tstate->curexc_value;
  PyObject *p_trace = tstate->curexc_traceback;

  tstate->curexc_type = NULL;
  tstate->curexc_value = NULL;
  tstate->curexc_traceback = NULL;

  std::vector<PyObject*> retval;
  retval.push_back(p_type);
  retval.push_back(p_val);
  retval.push_back(p_trace);

  return(retval);
}


std::string PythonLanguage::PyException::what()
{
  if( PyErr_Occurred() )
    PyErr_Print();
  
  return("");
}
*/
